/*
 * Creates a Hook with "priority", a double, and "func",
 * the function the hook will execute. params and parent
 * are optional.
 */

function Hook(priority, func, params) {
	var parameters = params || [];
	this.priority = priority;
	this.execute  = function() {
		func(parameters);
	}
}





/*
 * Creates a HookList, a list of Hooks ordered by priority.
 */
function HookList() {
	this.hooks  = [];
	this.length = 0;
}

/*
 * Adds "hook" to the HookList, and orders by priority.
 */
HookList.prototype.add = function(func, priority) {
	var p = priority || 1;
	this.length++;
	this.hooks.splice(this.__add_helper__(p), 0, func);
}

// Helper function for HookList.add(hook).
HookList.prototype.__add_helper__ = function(priority) {
	var low  = 0;
	var high = this.hooks.length;
	// binary search
	while (low < high) {
		var middle = Math.floor((low + high) / 2);
		if (priority < this.hooks[middle].priority)
			high = middle;
		else
			low = middle + 1;
	}
	return low;
}

/*
 * Adds hooks from a given list, and orders by priority.
 */
HookList.prototype.addHooks = function(hooks) {
	for (var i in hooks)
		this.addHook(hooks[i]);
}
/*
 * Removes "hook" from the HookList.
 */
HookList.prototype.remove = function(hook) {
	var i = 0;
	while (i < this.hooks.length) {
		if (hook == this.hooks[i]) {
			this.hooks.splice(i, 1);
			this.length--;
		} else
			i++;
	}
}

/*
 * Executes all hooks, ordered by priority.
 * Hooks that return true are removed.
 */
HookList.prototype.execute = function() {
	var i = 0;
	while (i < this.hooks.length) {
		var del = this.hooks[i]();
		if (del === true) {
			this.hooks.splice(i, 1);
			this.length--;
		} else
			i++;
	}
}