function SpriteImage(spritesetpath) {
	this.spriteset    = LoadSpriteset(spritesetpath);
	
	this.dir          = 0;
	this.framesNeeded = 0;
	this.currentFrame = 0;
	this.stopped      = false;
	this.looping      = true;
}

/**
 * @param x the x (in pixels) on the screen you want to blit it to.
 * @param y the y (in pixels) on the screen you want to blit it to.
 */
SpriteImage.prototype.blit = function(x, y) {
	this.spriteset.images[this.spriteset.directions[this.dir].frames[this.currentFrame].index].blit(x, y);
}

/**
 * @param x the x (in pixels, from the top-left of the person's BASE) on the screen you want to blit it to.
 * @param y the y (in pixels, from the top-left of the person's BASE) on the screen you want to blit it to.
 */
SpriteImage.prototype.blitFromBase = function(x, y) {
	this.blit(x - this.spriteset.base.x1, y - this.spriteset.base.y1);
}

/**
 * @param x the x (in pixels, on the map) on the screen you want to blit it to.
 * @param y the y (in pixels, on the map) on the screen you want to blit it to.
 */
SpriteImage.prototype.blitToMap = function(x, y) {
	this.blit(MapToScreenX(0, x), MapToScreenY(0, y));
}

/**
 * @param x the x (in pixels, on the map, from the top-left of the person's BASE) on the screen you want to blit it to.
 * @param y the y (in pixels, on the map, from the top-left of the person's BASE) on the screen you want to blit it to.
 */
SpriteImage.prototype.blitToMapFromBase = function(x, y) {
	this.blitToMap(x - this.spriteset.base.x1, y - this.spriteset.base.y1);
}

/**
 * Updates the frames:
 *    --if the # of frames passed >= current frame's delay, go to the next frame of the direction
 *        loops back to beginning of direction unless this.looping == false.
 *    --if the programmer has called .stop(), nothing happens.
 */
SpriteImage.prototype.update = function() {
	if(!this.stopped) {
		var frames = this.spriteset.directions[this.dir].frames;
		
		if(this.framesNeeded >= frames[this.currentFrame].delay) {
			if (this.looping) {
				this.framesNeeded = 0;
				this.currentFrame = (this.currentFrame + 1) % frames.length;
			}
		} else {
			this.framesNeeded++;
		}
	}
}

/**
 * Will not do anything if dirName is an invalid direction. Has O(n) complexity.
 * @param dirName the name of the direction you want the SpriteImage to face when blitted.
 */
SpriteImage.prototype.setDir = function(dirName) {
	var index = this.spriteset.directions.length;
	while(--index >= 0) {
		if(this.spriteset.directions[index].name == dirName) {
			this.dir = index;
			this.reset();
		}
	}
}

SpriteImage.prototype.getDir = function() {
	return this.spriteset.directions[this.dir].name;
}

/**
 * When called, the SpriteImage will no longer animate.
 */
SpriteImage.prototype.stop = function() {
	this.stopped = true;
}

/**
 * When called, the SpriteImage will continue animating.
 */
SpriteImage.prototype.resume = function() {
	this.stopped = false;
}

/**
 * When called, the SpriteImage's frames are reset to the first frame.
 */
SpriteImage.prototype.reset = function() {
	this.framesNeeded = 0;
	this.currentFrame = 0;
}

/**
 * Stops the SpriteImage from looping.
 */
SpriteImage.prototype.stopLoop = function() {
	this.looping = false;
}

/**
 * Resumes looping the SpriteImage.
 */
SpriteImage.prototype.resumeLoop = function() {
	this.looping = true;
}