const LOC_BOTTOM = 0;
const LOC_TOP    = 1;
const LOC_MIDDLE = 2;

// Be sure this is included after Scenario.js is included.
function Letter(font, asciiChar, x, y) {
	this.img = font.getCharacterImage(asciiChar);
	this.chara = asciiChar;
	this.x = x || 0;
	this.y = y || 0;
	
	// Letter-specific effects.
	this.color = CreateColor(255, 255, 255); // the default color of the text.
	this.shake = false; // if true, this letter will shake
}

Letter.prototype.blit = function(x, y) {
	// The letter (and all effects) is blitted here.
	if(this.shake)
		this.img.blitMask(this.x + x + Math.random() * 2, this.y + y + Math.random() * 2, this.color);
	else
		this.img.blitMask(this.x + x, this.y + y, this.color);
}


/*** MESSAGE ***/

CutScene.prototype.message = function(speaker, location) {
	var Instruction = {};
	Instruction.State = {};
	// You can edit any of these.
	Instruction.State.Font        = LoadFont("dialogue.rfn");
	Instruction.State.WindowStyle = LoadWindowStyle("black.rws");
	Instruction.State.Arrow       = LoadImage("MsgDownArrow.png");
	Instruction.State.InputKey    = GetPlayerKey(PLAYER_1, PLAYER_KEY_A);
	Instruction.State.InputAtLine = 3;
	Instruction.State.LineStart   = 1; // This is the offset for InputAtLine.
	Instruction.State.MaxLines    = 4;
	Instruction.State.TextSpeed   = 1;
	Instruction.State.Margin      = { top: 4, right: 4, bottom: 4, left: 4 };
	Instruction.State.Padding     = { top: 6, right: 6, bottom: 6, left: 6 };
	
	// Internals. Don't touch.
	Instruction.State.Pages = [];
	if (typeof location == "string") {
		for (var i = 1; i < arguments.length; i++)
			Instruction.State.Pages.push(arguments[i]);
		location = LOC_BOTTOM;
	} else {
		for (var i = 2; i < arguments.length; i++)
			Instruction.State.Pages.push(arguments[i]);
	}
	
	Instruction.State.Location     = location;
	Instruction.State.HasPressed   = false;
	Instruction.State.CurrentIndex = -1;
	Instruction.State.CurrentPage  = 0;
	Instruction.State.TargetOffset = 0;
	Instruction.State.Offset       = 0;
	Instruction.State.TimeElapsed  = -1;
	Instruction.State.Delay        = 0;
	Instruction.State.Mode         = "pull-up";
	Instruction.State.DrawnHeight  = 0;
	Instruction.State.WaitForInput = false;
	
	Instruction.State.SetDefaults = function() {
		// Undo one-time effects on the entire textbox.
		this.Delay = 0;
		this.TextSpeed = 1;
	};
	
	// Letter effects are applied to the letters beforehand.
	Instruction.State.ApplyLetterEffect = function(tag, letter) {
		switch(tag.name) {
			case "red":
				letter.color = CreateColor(255, 128, 128);
				break;
			case "blue":
				letter.color = CreateColor(128, 196, 255);
				break;
			case "gold":
				letter.color = CreateColor(255, 204, 0);
				break;
			case "green":
				letter.color = CreateColor(96, 255, 96);
				break;
			case "shake":
				letter.shake = true;
				break;
			default:
				return false;
		}
		return true;
	};
	
	// These effects affect the message as a whole, like delays, text-speed, etc.
	// They use the Instruction's State, which is more global than a single letter.
	Instruction.State.ApplyMessageEffect = function(tag, State) {
		switch(tag.name) {
			case "delay":
				State.Delay = tag.getAttr(200);
				break;
			case "speed":
				State.TextSpeed = tag.getAttr(2);
				break;
			case "input":
				State.WaitForInput = true;
				break;
			case "skip":
				State.CurrentIndex = tag.endIndex;
				break;
			case "exit":
				State.TurnPage();
				break;
			default:
				return false;
		}
		return true;
	};
	
	Instruction.State.ApplyEffects = function(tags, index, letter) {
		var i = tags.length;
		while(--i >= 0) {
			var tag = tags[i];
			if(index >= tag.startIndex && index < tag.endIndex) {
				if (letter != null)
					this.ApplyLetterEffect(tag, letter);
				else
					this.ApplyMessageEffect(tag, this);
			}
		}
	};
	
	Instruction.State.AdjustLocation = function() {
		this.Width  = GetScreenWidth() - this.Margin.left - this.Padding.left - this.Padding.right - this.Margin.right;
		this.Height = this.MaxLines * this.Font.getHeight();
		this.X      = this.Margin.left + this.Padding.left;
		
		if (this.Location == LOC_TOP) {
			this.Y = this.Margin.top + this.Padding.top;
		} else if(this.Location == LOC_MIDDLE) {
			this.Y = (GetScreenHeight() - this.Height - this.Padding.top - this.Margin.top) * 0.5;
		} else {
			this.Y =  GetScreenHeight() - this.Height - this.Padding.bottom - this.Margin.bottom;
		}
	};
	
	Instruction.State.CreateLetters = function(page, tags) {
		var letterArr  = [];
		var tagRegex   = /<.+?>/g;
		var textNoTags = page.replace(tagRegex, "");
		var textArray  = this.Font.wordWrapString(textNoTags, this.Width);
		
		// find indices of all necessary \ns.
		var prev = 0;
		var newlineAt = map(textArray, function(line) {
			prev += line.length;
			return prev;
		});
		// now remove all line breaks from textNoTags.
		textNoTags = textNoTags.replace(/[\r\n]/g, "");
		
		var index = -1;
		var lineStart = this.LineStart;
		var x = 0;
		var y = 0;
		
		var countLines = 0; // count # of lines so far.
		
		while(++index < textNoTags.length) {
			// add newlines as needed
			while(newlineAt.length > 0 && index == newlineAt[0]) {
				x = 0;
				y += this.Font.getHeight();
				countLines++;
				newlineAt.shift();
			}
			var letter = new Letter(this.Font, textNoTags.charCodeAt(index), x, y);
			this.ApplyEffects(tags, index, letter);
			letterArr.push(letter);
			
			x += this.Font.getStringWidth(textNoTags[index]);
		}
		
		return letterArr;
	};
	
	Instruction.State.ParseTags = function(str) {
		// line breaks cause a problem here because it counts as a character.
		// so we delete it for the time being.
		var text = str.replace(/[\r\n]/g, "");
		
		var tags              = [];
		var unclosedTagStack  = [];
		var tagRegex          = /<.+?>/g;
		var openTagRegex      = /^[^\/]+$/;
		var closeTagRegex     = /^\/[^\/]+$/;
		var selfCloseTagRegex = /^[^\/]+\/$/;
		var attributeRegex    = /^([A-Za-z]+)=(\w+)/g;
		var tagNameRegex      = /[A-Za-z]+/;
		var offset = 0;
		var myArray;
		
		function parseAttribute(tag) {
			var arr = tag.split("=");
			if(arr.length > 0)
				return arr[1];
			return undefined;
		}
		function trimWhitespace(str) {
			return str.replace(/^\s+/, "").replace(/\s+$/, "").replace(/\s+/, " ");
		}
		
		// find a tag in the text, and...
		while((myArray = tagRegex.exec(text)) !== null) {
			var tag     = trimWhitespace(myArray[0].substring(1, myArray[0].length - 1));
			var tagName = "" + tagNameRegex.exec(tag); // "" + because exec returns an object.
			offset     += myArray[0].length;
			if(openTagRegex.test(tag)) {
				unclosedTagStack.push({
					startIndex: tagRegex.lastIndex - offset,
					name:       tagName,
					attr:       parseAttribute(tag)
				});
			} else if(closeTagRegex.test(tag)) {
				if(unclosedTagStack.length > 0) {
					// pop! goes the weasel. get rid of closed tag on top of stack.
					var first   = unclosedTagStack.pop();
					if(tagName == first.name) {
						// now add appropriate effect.
						tags.unshift({
							startIndex: first.startIndex,
							endIndex:   tagRegex.lastIndex - offset,
							name:       tagName,
							attr:       first.attr,
							getAttr:    function(defaultValue) {
								if(this.attr !== undefined) {
									if(typeof defaultValue == "number")
										return parseFloat(this.attr);
									return this.attr;
								}
								return defaultValue;
							}
						});
					} else {
						// closed improperly!
						Abort(myArray[0] + ": You are closing \"" + tagName + "\" when you should be closing \"" + first.name + "\".");
					}
				} else {
					// closed improperly!
					Abort(myArray[0] + ": You are closing a tag that was never opened.");
				}
			} else if(selfCloseTagRegex.test(tag)) {
				// get rid of trailing '/' and whitespace
				tag = trimWhitespace(tag.substring(0, tag.length - 1));
				tags.unshift({
					startIndex: tagRegex.lastIndex - offset,
					endIndex:   tagRegex.lastIndex - offset + 1,
					name:       tagName,
					attr:       parseAttribute(tag),
					getAttr:    function(defaultValue) {
						if(this.attr !== undefined) {
							if(typeof defaultValue == "number")
								return parseFloat(this.attr);
							return this.attr;
						}
						return defaultValue;
					}
				});
			} else {
				// Not a valid tag! The default ones are <open> </close> <selfclose/>
				Abort(myArray[0] + ": This is not a valid tag. Please fix.");
			}
		}
		
		return tags;
	};
	
	Instruction.State.IncrementIndex = function() {
		var reps = (IsKeyPressed(this.InputKey)) ? 3 : 1;
		var formerIndex = index = this.CurrentIndex;
		var keepGoing   = true;
		do {
			this.CurrentIndex += this.TextSpeed;
			index = this.CurrentIndex;
			forEach(this.Pages[this.CurrentPage].tags, function(tag) {
				if (tag.startIndex == Math.floor(index + 1))
					keepGoing = false;
			});
		} while (--reps > 0 && keepGoing && Math.floor(this.CurrentIndex) <= this.Pages[this.CurrentPage].text.length);
		// only update time if you have increased one index.
		if (Math.floor(this.CurrentIndex) > Math.floor(formerIndex))
			this.TimeElapsed = GetTime();
	};
	
	Instruction.State.IsInputEnded = function() {
		var isPressed = IsKeyPressed(this.InputKey);
		if (!this.HasPressed) {
			if (!isPressed)
				this.HasPressed = true;
		} else if (isPressed) {
			this.HasPressed = false;
			return true;
		}
		return false;
	};
	
	Instruction.State.TurnPage = function() {
		this.CurrentPage++;
		this.CurrentIndex = -1;
		this.Offset = 0;
		this.TargetOffset = 0;
	};
	
	Instruction.start = function(State) {
		State.TimeElapsed = GetTime();
		State.AdjustLocation();
		
		State.Pages = map(State.Pages, function(words) {
			words = "<gold>" + speaker + "</gold>:\n" + words + "<input/>";
			words = words.replace(/,/g, ",<delay=200/>");
			words = words.replace(/((?:\?|!|\.)+(?=\s))/g, "$1<delay=400/>");
			
			var tags = State.ParseTags(words);
			var text = State.CreateLetters(words, tags);
			// Now that we've applied letter-only tags, delete them.
			var l = new Letter(State.Font, 32);
			tags = filter(tags, function(tag) {
				return !State.ApplyLetterEffect(tag, l);
			});
			return {
				tags: tags,
				text: text
			};
		});
		return true;
	};
	
	Instruction.render = function(State) {
		var y = State.Y;
		if (State.Location != LOC_TOP)
			y += State.Height - State.DrawnHeight;
		var height = State.DrawnHeight;
		
		State.WindowStyle.drawWindow(State.X - State.Padding.left,
									 y - State.Padding.top,
									 State.Width  + State.Padding.left + State.Padding.right,
									 height + State.Padding.top  + State.Padding.bottom);
		SetClippingRectangle(State.X, y, State.Width, height);
		if (State.CurrentPage < State.Pages.length) {
			var letters = State.Pages[State.CurrentPage].text;
			for (var i = 0; i <= Math.min(State.CurrentIndex, letters.length - 1); i++)
				letters[i].blit(State.X, y - State.Offset);
		}
		SetClippingRectangle(0, 0, GetScreenWidth(), GetScreenHeight());
		if (State.WaitForInput) {
			State.Arrow.blit(State.X + State.Width - State.Arrow.width,
							 y + height - State.Arrow.height - 2
									 + Math.sin((GetTime() - State.TimeElapsed)>>7));
		}
	};
	
	Instruction.update = function(State) {
		switch (State.Mode) {
			case "pull-up":
				State.DrawnHeight = easeIn(0, State.Height, (GetTime() - State.TimeElapsed) / 250);
				if (State.DrawnHeight >= State.Height) {
					State.DrawnHeight = State.Height;
					State.Mode = "text";
				}
				return true;
			case "text":
				var page = State.Pages[State.CurrentPage];
				var currentLetter = page.text[Math.floor(State.CurrentIndex)];
				
				// Offsets
				if (State.Offset < State.TargetOffset) State.Offset++;
				if (State.CurrentIndex >= 0 && currentLetter.y >= State.MaxLines * State.Font.getHeight() + State.TargetOffset)
					State.TargetOffset += State.Font.getHeight();
				
				if (GetTime() > State.TimeElapsed + State.Delay && !State.WaitForInput) {
					State.IncrementIndex();
					State.SetDefaults();
					State.ApplyEffects(page.tags, Math.floor(State.CurrentIndex + 1));
				}
				if (State.CurrentIndex >= page.text.length)  State.TurnPage();
				if (State.CurrentPage >= State.Pages.length)
					State.Mode = "pull-down";
				return true;
			case "pull-down":
				State.DrawnHeight = easeOut(State.Height, 0, (GetTime() - State.TimeElapsed) / 250);
				return State.DrawnHeight > 0;
		}
	};
	
	Instruction.handleInput = function(State) {
		if (State.WaitForInput && State.IsInputEnded()) {
			State.WaitForInput = false;
		}
	};
	
	this.enqueue(Instruction);
}